import hdsubnetfinder.subnetwork.network_util as util


class SubNetworkFinder():
    """
    This service gets an annotated node network
    typically derived from an uploaded TSV
    and uses a reference network to find a
    relevant subnetwork.  The annotations are
    applied to the subnetwork and it is returned
    as CX network as the result of the job
    """

    def __init__(self, kernel, network):
        """
        Sub network finder is always associated with a network and pre-computed kernel.

        :param kernel:
        :param network: Network as triples
        :return:
        """
        if kernel is None:
            raise ValueError('Kernel is required.')
        if network is None:
            raise ValueError('Network is required.')

        self.__kernel = kernel
        self.__network = network

    def get_sub_network(self, identifiers):
        query_vector = util.query_vector(identifiers, self.__kernel.labels)
        diffused = self.__kernel.diffuse(query_vector)
        filtered = util.filter_sif(self.__network, diffused)

        # Return result in CX format.
        return util.sif2cx(filtered, scores=diffused)



from array import array
import time
import urllib2
import logging
import csv
import numpy as np

from scipy.sparse import coo_matrix, csc_matrix
from scipy.sparse.linalg import expm

from kernel import Kernel

logging.basicConfig(level=logging.DEBUG)


class KernelGenerator:

    def create_kernel_from_file(self, pre_computed_kernel_url):
        """
        Input:
            pre_computed_kernel_file - filename of tab delemited kernel file,
                                        as made by KernelGenerator
        """
        ker = []
        start = True
        labels = list()

        for line in csv.reader(urllib2.urlopen(pre_computed_kernel_url), delimiter='\t'):
            if start:
                labels = line[1:]
                start = False
            else:
                ker.append([float(x) for x in line[1:]])

        kernel = csc_matrix(np.asarray(ker))
        return Kernel(kernel=kernel, labels=labels, index2node=None)

    def create_kernel(self, network_url, time_t=0.1):
        start = time.clock()

        edges, nodes, node_out_degrees = self.__parse_net(network_url)

        num_nodes = len(nodes)
        node_order = list(nodes)

        index2node = {}
        node2index = {}

        logging.debug('Network source: ' + str(network_url))
        logging.debug('# of Nodes: ' + str(num_nodes))
        logging.debug('# of Edges: ' + str(len(edges)))

        for i in range(0, num_nodes):
            index2node[i] = node_order[i]
            node2index[node_order[i]] = i

        # construct the diagonals
        # SCIPY uses row and column indexes to build the matrix
        # row and columns are just indexes: the data column stores 
        # the actual entries of the matrix
        row = array('i')
        col = array('i')
        data = array('f')

        # build the diagonals, including the out-degree 
        for i in range(0, num_nodes):
            # diag entries: out degree
            degree = 0
            if index2node[i] in node_out_degrees:
                degree = node_out_degrees[index2node[i]]
                # append to the end
            # array object: first argument is the index, the second is the data value
            # append the out-degree to the data array
            data.insert(len(data), degree)
            # build the diagonals
            row.insert(len(row), i)
            col.insert(len(col), i)

            # add off-diagonal edges
        for i in range(0, num_nodes):
            for j in range(0, num_nodes):
                if i == j:
                    continue
                if (index2node[i], index2node[j]) not in edges:
                    continue
                # append index to i-th row, j-th column
                row.insert(len(row), i)
                col.insert(len(col), j)
                # -1 for laplacian: i.e. the negative of the adjacency matrix 
                data.insert(len(data), -1)

        # Build the graph laplacian: the CSC matrix provides a sparse matrix format
        # that can be exponentiated efficiently
        l = coo_matrix((data, (row, col)), shape=(num_nodes, num_nodes)).tocsc()

        end = time.clock()
        logging.debug('Data preparation done in ' + str(end - start) + ' sec.')

        # this is the matrix exponentiation calculation.
        # Uses the Pade approximiation for accurate approximation.
        # Computationally expensive.
        # O(n^2), n= # of features, in memory as well. 
        start = time.clock()
        kernel = expm(-time_t * l)
        end = time.clock()

        logging.debug('expm done in ' + str(end - start) + ' sec.')

        return Kernel(kernel=kernel, labels=node_order, index2node=index2node)

    def __parse_net(self, network_url):
        """
        Parse .sif network, using just the first and third columns
        to build an undirected graph. Store the node out-degrees
        in an index while we're at it. 
        """
        edges = set()
        nodes = set()
        degrees = {}

        for line in urllib2.urlopen(network_url):
            parts = line.rstrip().split()

            # Remove orphan node
            if len(parts) != 3:
                continue

            source = parts[0]
            target = parts[2]

            # if inputting a multi-graph, skip this
            if (source, target) in edges:
                continue

            # Remove self-loop
            if source == target:
                continue

            edges.add((source, target))
            edges.add((target, source))

            nodes.add(source)
            nodes.add(target)

            if source not in degrees:
                degrees[source] = 0
            if target not in degrees:
                degrees[target] = 0

            degrees[source] += 1
            degrees[target] += 1

        return edges, nodes, degrees



# -*- coding: utf-8 -*-


def write_kernel(target_kernel, output_file):
    """
    Serialize the kernel to the supplied output file
    """
    index2node = target_kernel.index2node

    if index2node is None:
        raise ValueError('Cannot serialize Kernel created from pre-computed file.')

    kernel = target_kernel.kernel
    labels = target_kernel.labels

    output_stream = open(output_file, 'w')

    cx = kernel.tocoo()
    edges = {}
    for i, j, v in zip(cx.row, cx.col, cx.data):
        a = index2node[i]
        b = index2node[j]
        edges[(a, b)] = str(v)

    # iterate through rows
    # sort labels in alphabetical order

    output_stream.write("Key\t" + "\t".join(sorted(labels)) + "\n")

    for nodeA in sorted(labels):
        printstr = nodeA
        # through columns
        for nodeB in sorted(labels):
            if (nodeA, nodeB) in edges:
                printstr += "\t" + edges[(nodeA, nodeB)]
            else:
                printstr += "\t0"

        output_stream.write(printstr + "\n")

    output_stream.close()



import os
import unittest
import pandas as pd

from hdsubnetfinder.kernel.kernel_generator import KernelGenerator
from hdsubnetfinder.kernel.kernel import Kernel
import hdsubnetfinder.kernel.kernel_util as util


class KernelGeneratorTests(unittest.TestCase):

    def test_kernel_generator_network(self):

        def run(network):
            generator = KernelGenerator()

            # First, read network file as table
            small_net_df = pd.read_csv(network, sep='\t',
                                       names=['source', 'interaction', 'target'])
            print(small_net_df.shape)

            sources = small_net_df['source']
            targets = small_net_df['target']
            merged = sources.append(targets)
            print(len(merged.unique()))

            # Calculate kernel
            small_kernel = generator.create_kernel(network)

            # Serialize file
            util.write_kernel(small_kernel, 'kernel_out.txt')

            # Get kernel as big string
            self.assertIsNotNone(small_kernel)
            self.assertTrue(isinstance(small_kernel, Kernel))

            self.assertIsNotNone(small_kernel.kernel)
            self.assertIsNotNone(small_kernel.labels)

            df = pd.read_csv('kernel_out.txt', sep='\t', index_col=0)

            self.assertIsNotNone(df)
            print(df.shape)
            # self.assertEqual(generator.num_nodes, df.shape[0])
            # self.assertEqual(generator.num_nodes, df.shape[1])

        print('\n---------- Kernel Generator tests start -----------\n')

        # Small network (takes about 10 seconds to compute)
        small_network_file = 'resource/small_network.sif'
        net_file = os.path.abspath(os.path.dirname(__file__)) + '/' + small_network_file
        small_network_url = 'file://' + net_file
        #small_network = 'https://s3-us-west-2.amazonaws.com/ci-service-data/small.sif'

        # Medium size network (takes about 5 minutes)
        med_network = 'https://s3-us-west-2.amazonaws.com/ci-service-data/yeastHighQuality.sif'

        run(small_network_url)
        #run(med_network)

        print('\n---------- Generator tests finished! -----------\n')



# -*- coding: utf-8 -*-


class Kernel:

    def __init__(self, kernel=None, labels=None, index2node=None):
        if labels is None:
            raise ValueError('labels are required.')

        if kernel is None:
            raise ValueError('Kernel is required.')

        self.labels = labels
        self.kernel = kernel
        self.index2node = index2node

    def kernel_multiply_one(self, vector):
        """
            Multiply the specified kernel by the supplied input heat vector.

            Input:
                vector: A hash mapping gene labels to floating point values
                kernel: a single index for a specific kernel

            Returns:
                A hash of diffused heats, indexed by the same names as the
                input vector
        """

        # Have to convert to ordered array format for the input vector
        array = []
        for label in self.labels:
            # Input heats may not actually be in the network.
            # Check and initialize to zero if not
            if label in vector:
                array.append(vector[label])
            else:
                array.append(0)

        # take the dot product
        value = self.kernel * array

        # Convert back to a hash and return diffused heats
        return_vec = {}
        idx = 0
        for label in self.labels:
            return_vec[label] = float(value[idx])
            idx += 1

        return return_vec

    def diffuse(self, vector, reverse=False):
        """
        Diffuse input heats over the set of kernels, add to this object

        Input:
            {
                'gene1': float(heat1)
                'gene2' : float(heat2)
                ...
            }

        Returns:
            Diffused heat vector
        """

        return self.kernel_multiply_one(vector)



# -*- coding: utf-8 -*-
import csv
import urllib2
import operator


def read_sif(file_url):
    """
    Read a SIF network file and create triples as a list

    :param file_url: Location of the input SIF file
    :return: list of strings (source, interaction, target)
    """
    triples = []
    reader = csv.reader(urllib2.urlopen(file_url), delimiter='\t')

    for line in reader:
        triples.append(line)

    return triples


def filter_sif(triples, scores, desired_nodes=30):
    scores_made_the_cut = sorted(
        scores.items(),
        key=operator.itemgetter(1),
        reverse=True)[0:(desired_nodes)]

    nodes_made_the_cut = [i[0] for i in scores_made_the_cut]
    edges_made_the_cut = []

    for tr in triples:
        if (tr[0] in nodes_made_the_cut) and (tr[2] in nodes_made_the_cut):
            edges_made_the_cut.append(tr)

    return edges_made_the_cut


def sif2cx(triples, scores=None):
    cx_out = []
    identifier_to_node_id_map = {}
    id_counter = 1000

    for tr in triples:
        subj_node_id = identifier_to_node_id_map.get(tr[0])
        obj_node_id = identifier_to_node_id_map.get(tr[2])

        if not subj_node_id:
            subj_node_id = "_" + str(id_counter)
            id_counter = id_counter + 1
            identifier_to_node_id_map[tr[0]] = subj_node_id
            cx_out.append({"nodes": [{"@id": subj_node_id}]})
            cx_out.append({"nodeIdentities": [{"node": subj_node_id, "represents": tr[0]}]})

        if not obj_node_id:
            obj_node_id = "_" + str(id_counter)
            id_counter = id_counter + 1
            identifier_to_node_id_map[tr[2]] = obj_node_id
            cx_out.append({"nodes": [{"@id": obj_node_id}]})
            cx_out.append({"nodeIdentities": [{"node": obj_node_id, "represents": tr[2]}]})

        edge_section = {
            "edges": [
                {
                    "source": subj_node_id, "@id": "_" + str(id_counter),
                    "target": obj_node_id
                }
            ],
            "edgeIdentities": [
                {
                    "edge": "_" + str(id_counter),
                    "relationship": tr[1]
                }
            ]
        }
        cx_out.append(edge_section)
        id_counter = id_counter + 1

    if scores:
        for node_name, value in scores.iteritems():
            node_id = identifier_to_node_id_map.get(node_name)
            if node_id:
                node_section = {
                    "elementProperties": [
                        {
                            "node": node_id,
                            "property": "subnet_finder_score",
                            "value": value
                        }
                    ]
                }
                cx_out.append(node_section)

    return cx_out


def query_vector(query, labels):
    out = {}
    weight_sum = 0.0
    for i in labels:
        if i in query:
            out[i] = 1.0
            weight_sum += 1.0
        else:
            out[i] = 0.0

    for i in labels:
        out[i] = out[i] / weight_sum

    return out



# -*- coding: utf-8 -*-
import unittest
import pandas as pd
import json

small_network = 'https://s3-us-west-2.amazonaws.com/ci-service-data/small.sif'


class NetworkUtilTests(unittest.TestCase):

    def test_read_sif(self):
        import hdsubnetfinder.subnetwork.network_util as util
        df = pd.read_csv(small_network, sep='\t', names=['source', 'interaction', 'target'])
        print(df.shape)

        triples = util.read_sif(small_network)
        self.assertIsNotNone(triples)
        self.assertEqual(df.shape[0], len(triples))

    def test_sif2cx(self):
        import hdsubnetfinder.subnetwork.network_util as util
        triples = util.read_sif(small_network)
        cx = util.sif2cx(triples)
        print(json.dumps(cx, indent=4))

        self.assertEqual(list, type(cx))
        # Add more tests here...



# -*- coding: utf-8 -*-

import unittest
import os
from hdsubnetfinder.kernel.kernel_generator import KernelGenerator

KERNEL_LOCATION = 'resource/kernel.txt'
k_file = os.path.abspath(os.path.dirname(__file__)) + '/' + KERNEL_LOCATION
SAMPLE_KERNEL_URL = 'file://' + k_file

NETWORK_LOCATION = 'resource/small_network.sif'
net_file = os.path.abspath(os.path.dirname(__file__)) + '/' + NETWORK_LOCATION
NETWORK_URL = 'file://' + net_file
print(NETWORK_URL)


class SubNetworkFinderTests(unittest.TestCase):

    def test_find_sub_network(self):
        print('\n---------- Sub Network Finder tests start -----------\n')

        from hdsubnetfinder.subnetwork.sub_network_finder import SubNetworkFinder
        import hdsubnetfinder.subnetwork.network_util as util


        # Create kernel from pre-computed kernel file
        generator = KernelGenerator()
        small_kernel = generator.create_kernel_from_file(pre_computed_kernel_url=SAMPLE_KERNEL_URL)
        small_sif = util.read_sif(file_url=NETWORK_URL)

        finder = SubNetworkFinder(network=small_sif, kernel=small_kernel)

        identifiers = ["NRAS", "KRAS", "MAPK1"]

        result = finder.get_sub_network(identifiers)
        self.assertIsNotNone(result)

        print(len(result))

        self.assertEqual(list, type(result))

        print('\n---------- finder tests finished! -----------\n')

    def test_find_sub_network2(self):
        print('\n---------- Sub Network Finder tests 2 start -----------\n')

        from hdsubnetfinder.subnetwork.sub_network_finder import SubNetworkFinder
        import hdsubnetfinder.subnetwork.network_util as util
        from hdsubnetfinder.kernel.kernel_generator import KernelGenerator

        generator = KernelGenerator()
        small_kernel = generator.create_kernel(network_url=NETWORK_URL)
        small_sif = util.read_sif(file_url=NETWORK_URL)

        finder = SubNetworkFinder(network=small_sif, kernel=small_kernel)

        identifiers = ["NRAS", "KRAS", "MAPK1"]

        result = finder.get_sub_network(identifiers)
        self.assertIsNotNone(result)
        print(len(result))
        self.assertEqual(list, type(result))

        print('\n---------- finder tests2 finished! -----------\n')

    # def test_find_sub_network3(self):
    #     from hdsubnetfinder.subnetwork.sub_network_finder import SubNetworkFinder
    #     import hdsubnetfinder.subnetwork.network_util as util
    #     from hdsubnetfinder.kernel.kernel_generator import KernelGenerator
    #
    #     generator = KernelGenerator()
    #     med_network = 'https://s3-us-west-2.amazonaws.com/ci-service-data/yeastHighQuality.sif'
    #     med_kernel = generator.create_kernel(network_url=med_network)
    #     med_sif = util.read_sif(file_url=med_network)
    #
    #     finder = SubNetworkFinder(network=med_sif, kernel=med_kernel)
    #
    #     identifiers = ["YNL031C", "YNL145W"]
    #
    #     result = finder.get_sub_network(identifiers)
    #     self.assertIsNotNone(result)
    #     print(len(result))
    #     self.assertEqual(list, type(result))
    #
    #     print('\n---------- finder tests2 finished! -----------\n')



# -*- coding: utf-8 -*-
import unittest
import os

KERNEL_LOCATION = 'resource/kernel.txt'
k_file = os.path.abspath(os.path.dirname(__file__)) + '/' + KERNEL_LOCATION
SAMPLE_KERNEL_URL = 'file://' + k_file
print(SAMPLE_KERNEL_URL)


class KernelTests(unittest.TestCase):

    def test_pre_computed_kernel(self):
        from hdsubnetfinder.kernel.kernel_generator import KernelGenerator

        generator = KernelGenerator()
        small_kernel = generator.create_kernel_from_file(pre_computed_kernel_url=SAMPLE_KERNEL_URL)

        self.assertIsNotNone(small_kernel)
        self.assertIsNotNone(small_kernel.kernel)
        self.assertIsNotNone(small_kernel.labels)

        self.assertTrue(list, type(small_kernel.labels))
        self.assertTrue(list, type(small_kernel.kernel))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""

Setup script for hdsubnetfinder

To install, run:

python setup.py install

"""

from setuptools import setup, find_packages

setup(
    name='hdsubnetfinder',
    version='0.2.1',
    description='Heat Diffusion Sub-network Finder',
    long_description='Sub-network finder using heat diffusion simulation.',
    author='Daniel Carlin, Keiichiro Ono',
    author_email='kono@ucsd.edu',
    url='https://github.com/idekerlab/hdsubnetfinder',
    license='MIT License',
    install_requires=[
        'requests',
        'pyparsing',
        'numpy',
        'scipy'
    ],
    keywords=['bioinformatics', 'graph', 'network', 'cytoscape'],
    classifiers=[
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 2.7',
        'License :: OSI Approved :: MIT License',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Mathematics',
    ],
    test_suite='tests',
    packages=find_packages(),
    include_package_data=True
)
